#!/usr/bin/env perl
# /*---------------------------------------------------------------------------
#  * Copyright 2009 Igor Daniloff.
#  * 
#  * Following source code is the property of Doctor Web, Ltd.
#  * Use is subject to license terms.
#  * The reproduction in any form without written permission of
#  * Doctor Web, Ltd. and proper attribution is strongly prohibited.
#  * 
#  * Author: $Author: a.rikov $
#  * Date: $Date: 2011/08/15 09:27:58 $
#  * Revision: $Revision: 1.12.8.1 $
#  ----------------------------------------------------------------------------*/


use strict;
use warnings;


{
    package Exception::Base;
    use strict;
    use fields qw(error_str);
    use overload ('""' => \&stringify);


    sub new
    {
        my ($invocant, $error_str)     = @_;

        $invocant = ref $invocant if ref $invocant;
        my Exception::Base $self = fields::new($invocant);
        $self->{error_str} = $error_str;

        return $self;
    }

 
    sub stringify
    {
        my Exception::Base $self = shift;
        return $self->{error_str};
    }
}


{
    package Exception::IO;
    use base qw(Exception::Base);
}


{
    package Exception::Envelope;
    use base qw(Exception::Base);
}


{
    package Logger;
    use strict;
    use fields qw(log_level);


    use constant DEFAULT_LOG_LEVEL => 'debug';
    use constant LOG_LEVEL_LIST => { 
        quiet   => 0,
        error   => 1,
        warning => 2,
        info    => 3,
        debug   => 4,
    };


    sub new
    {
        my ($invocant, $log_level)     = @_;

        $invocant = ref $invocant if ref $invocant;
        my Logger $self = fields::new($invocant);

        if (defined $log_level) {
            $log_level = lc $log_level;
            $self->{log_level} = exists &LOG_LEVEL_LIST->{$log_level}
                                     ? &LOG_LEVEL_LIST->{$log_level}
                                     : &LOG_LEVEL_LIST->{&DEFAULT_LOG_LEVEL};
        } else {
            $self->{log_level} = &LOG_LEVEL_LIST->{&DEFAULT_LOG_LEVEL}
        }

        return $self;
    }


    sub log
    {
        my Logger $self = shift;
        my $log_level   = shift;

        die "Wrong log level: '$log_level'"
           if !defined $log_level || !exists &LOG_LEVEL_LIST->{$log_level = lc $log_level};
        return if &LOG_LEVEL_LIST->{$log_level} > $self->{log_level};
        printf "%s: [%s] %s\n", scalar localtime(), $log_level,
                                defined $_[0]
                                   ? ref $_[0] eq 'SCALAR'
                                       ? ${$_[0]}
                                       : $_[0]
                                   : '';
        return 1;
    }


}


{
    package WrapperSocket;
    use strict;
    use fields qw(logger socket timeout);

    use IO::Socket qw();
    use Errno      qw();

    use constant DEFAULT_TIMEOUT => 60;


    sub new
    {
       my $invocant = shift;
       my %args     = @_;

       my $peer_address  = $args{peer_address};
       my Logger $logger = $args{logger};
    
       $invocant = ref $invocant if ref $invocant;
 
       $logger->log('debug', 'Create socket object');
       $logger->log('debug', "Opening socket: peer_addr = $peer_address; type = " . &IO::Socket::SOCK_STREAM . ' .');
       my $socket = IO::Socket::UNIX->new(
           Peer     => $peer_address,
           Type     => &IO::Socket::SOCK_STREAM,
       )
           or die Exception::IO->new("Can't connect to '$peer_address': $!");
       $socket->autoflush(1);
    
       my WrapperSocket $self = fields::new($invocant);
       $self->{socket}   = $socket;
       $self->{timeout}  = $args{timeout} || &DEFAULT_TIMEOUT;
       $self->{logger}   = $logger;

       $self->receiveData(my $output); #clear socket from "hello" messages
    
       return $self;
    }


    DESTROY
    {
        my WrapperSocket $self = shift;
        my Logger $logger = $self->{logger};

        $logger->log('debug', 'Destroy socket object');

        $logger->log('debug', 'Shutdown socket.');
        $self->{socket}->shutdown(2);
    }


    sub sendData
    {
        my WrapperSocket $self = shift;
        my $data               = shift;

        my Logger $logger = $self->{logger};

        $logger->log('debug', 'Sending data to daemon: ' . $data);
        $self->{socket}->print($data, "\n")
            or die Exception::IO->new("Can not send data by socket: $!");
        return 1;
    }


    sub receiveData
    {
        my WrapperSocket $self = shift;

        my Logger $logger = $self->{logger};

        $logger->log('debug', 'Receive data from daemon.');

        my ($error, $result, $buffer);
        eval {
            local $SIG{ALRM} = sub {die Exception::IO->new('Can not read data from socket: timeout.')};
            alarm $self->{timeout};
            eval {
                while(1) {
                    unless(defined( $buffer = $self->{socket}->getline() )) {
                        if ($!) {
                            die Exception::IO->new("Can not get data by socket: $!");
                        } else {
                            $buffer = ''; #EOF
                        }
                    }
                    chomp($buffer);
                    $logger->log('debug', 'Received data: ' . $buffer);
                    last if $buffer eq '';
                    $_[0] .= $buffer;
                    $_[0] .= "\n";
                }
            };
            $error = $@ if $@;
            alarm 0;
        };
        alarm 0;
        $@ = $error if $error;
        if ($@) { die $@; }

        return $result;
    }
}


{
    package Envelope;
    use strict;
    use fields qw(logger eml_file env_file id from to infected_dir);

    sub new
    {
        my $invocant = shift;
        my %args     = @_;

        my Logger $logger = $args{logger};

        $invocant = ref $invocant if ref $invocant;

        $logger->log('debug', 'Create envelope object');
        my Envelope $self = fields::new($invocant);

        @$self{qw(infected_dir env_file)}  = @args{qw(infected_dir env_file)};
        $self->{infected_dir} .= '/' if  $self->{infected_dir} !~ /\/$/;
        $self->{logger} = $logger;

        return $self;
    }


    sub needToProcess
    {
        my Envelope $self = shift;
        my Logger $logger = $self->{logger};

        $logger->log('debug', "Check need to process file '$self->{env_file}' or not.");
        my $result = 1;
        unless(-f $self->getEmlFile()) {
            $logger->log('debug', "Envelope '$self->{env_file}' does not have message file.");
            $result = 0;
        }

        $logger->log('debug', $result ? "Need to process file '$self->{env_file}'" : "Not need to process file '$self->{env_file}'");
        return $result;
    }


    sub getEmlFile
    {
        my Envelope $self = shift;
        my Logger $logger = $self->{logger};

        $logger->log('debug', "Message file name for envelope '$self->{env_file}' has been requested.");

        if(defined $self->{eml_file}) {
            $logger->log('debug', "Message file name has already generated: $self->{eml_file}");
            return $self->{eml_file};
        }

        $logger->log('debug', "Need to generete message file name for file: $self->{env_file}");
        my $eml_file = $self->{env_file};
        $eml_file =~ s/\.envelope$//;
        $logger->log('debug', "Genereted message file name: $eml_file");
        return $self->{eml_file} = $eml_file;
    }


    sub getId
    {
        my Envelope $self = shift;
        my Logger $logger = $self->{logger};

        $logger->log('debug', "Id for envelope '$self->{env_file}' has been requested.");

        if(defined $self->{id}) {
            $logger->log('debug', "Id has been already generated: $self->{id}");
            return $self->{id};
        }

        $logger->log('debug', "Need to generete id for file: $self->{env_file}");
        my $id  = $self->getEmlFile();
        substr($id, 0, length $self->{infected_dir}) = '';
        $id =~ s/^\/+//;
        $logger->log('debug', "Genereted id: " . (defined $id ? $id : '<not defined>' ));
        return $self->{id} = $id;
    }


    sub getEnvFile
    {
        my Envelope $self = shift;
        my Logger $logger = $self->{logger};

        $logger->log('debug', "Envelope file name '$self->{env_file}' has been requested.");

        return $self->{env_file};
    }


    sub getAddresses
    {
        my Envelope $self = shift;
        my Logger $logger = $self->{logger};

        $logger->log('debug', "Treing to get addresses from file: $self->{env_file}");
        open(my $fh, q{<}, $self->{env_file}) or die Exception::Envelope->new("Can't open file '$self->{env_file}': $!");

        $logger->log('debug', "Getting sender adress.");
        $self->{from} = $self->getAddressFromFile($fh, 'sender');

        $logger->log('debug', "Reading number of recepients");
        my $rcpts_num;
        if(read($fh,$rcpts_num,4)!=4){
            close $fh;
            die Exception::Envelope->new("Can`t read number of rcpts in envelope file: $!");
        }
        $rcpts_num = unpack("N",$rcpts_num);
        $logger->log('debug', "Number of recepients is $rcpts_num");

        $logger->log('debug', "Getting addresses of recepients");
        my @to = ();my $rcpt;
        while($rcpts_num--){
            $rcpt = $self->getAddressFromFile($fh, 'recepient');
            push @to, $rcpt;
        }
        $self->{to} = \@to;

        close($fh);
        return 1;
    }


    sub getFrom
    {
        my Envelope $self = shift;
        my Logger $logger = $self->{logger};

        $logger->log('debug', "Sender address for envelope '$self->{env_file}' has been requested.");

        if(defined $self->{from}) {
            $logger->log('debug', "Sender address has been already generated: $self->{from}");
            return $self->{from};
        }

        $logger->log('debug', "Need to generete sender address for file: $self->{env_file}");
        $self->getAddresses();
        $logger->log('debug', "Genereted sender address: $self->{from}");
        return $self->{from};
    }


    sub getTo
    {
        my Envelope $self = shift;
        my Logger $logger = $self->{logger};

        $logger->log('debug', "Recepients addresses for envelope '$self->{env_file}' have been requested.");

        if(defined $self->{to}) {
            $logger->log('debug', "Recepients addresses have been already generated: " . (join ", ",@{$self->{to}}));
            return $self->{to};
        }

        $logger->log('debug', "Need to generete recepients addresses for file: $self->{env_file}");
        $self->getAddresses();
        $logger->log('debug', "Genereted recepients addresses: " . (join ", ",@{$self->{to}}));
        return $self->{to};
    }


    sub getAddressFromFile
    {
        my Envelope $self = shift;
        my ($fh, $type)   = @_;

        my Logger $logger = $self->{logger};

        $logger->log('debug', "Read pascal string from: $self->{env_file}.");
        my $data = $self->readPascalString($fh);

        $logger->log('debug', "Parse readed data.");
        $data = $1 if $data =~ /^\s*<\s*(.+?)\s*>\s*$/;
        if (!defined $data || $data =~ /^\s*$/) {
            #close $fh;
            #die Exception::Envelope->new("Find bad $type address in '$self->{env_file}': $data");
            $data = '';
        }
        $logger->log('debug', "Parsed data: $data");
        return $data;
    }


    sub readPascalString
    {
        my Envelope $self = shift;
        my $fh            = shift;

        my Logger $logger = $self->{logger};

        $logger->log('debug', "Reading string length.");
        my $num;
        if(read($fh,$num,4)!=4){
            close $fh;
            die Exception::Envelope->new("Can`t read 4 bytes for string length: $!");
        }
        $num = unpack("N",$num);
        $logger->log('debug', "String length: $num");

        $logger->log('debug', "Reading string.");
        my $data;
        if(read($fh,$data,$num)!=$num){
            close $fh;
            die Exception::Envelope->new("Can`t read $num bytes: $!");
        }
        return $data;
    }


    sub destroy
    {
        my Envelope $self = shift;
        my Logger $logger = $self->{logger};

        $logger->log('debug', "Destroing envelope $self->{env_file}");
        
        $logger->log('debug', "Delete file envelope $self->{env_file}");
        unlink $self->{env_file}
            or die Exception::Envelope->new("Can't delete file '$self->{env_file}':$!");
        return 1;
    }

}


use Fcntl           qw();
use File::Find      qw();
use Getopt::Long    qw();

use constant DEFAULT_LOCK_FILE            => '/var/drweb/spool/migrator.lock';
use constant DEFAULT_LOG_LEVEL            => 'info';
use constant DEFAULT_PATH_TO_INFECTED_DIR => '/var/drweb/infected';
use constant DEFAULT_SOCKET_PATH          => '/var/drweb/ipc/.ctl';

my $exit_flag = 0;

$SIG{__WARN__} = $SIG{__DIE__} = sub {!ref $_[0] && die Exception::Base->new($_[0])};
$SIG{INT}  = sub {$exit_flag = 1; $SIG{PIPE} = 'IGNORE'; print "\nCatched signal to quit.\n"};
$SIG{PIPE} = sub {$SIG{INT} = $SIG{PIPE} = 'IGNORE';die "Broken pipe. Exiting..."};


MAIN:
{
   my Logger $logger = Logger->new();

   my ($socket, $quit_flag, $start_time, $end_time);
   my ($processed_files, $skipped_files, $errors) = (0, 0, 0);
   $| = 1;

   eval{
       my ($help, $socket_path, $infected_dir, $log_level, $output, $lock_file);

       Getopt::Long::GetOptions(
           'h|help'           => \$help,
           's|socket-path=s'  => \$socket_path,
           'q|infected-dir=s' => \$infected_dir,
           'l|log-level=s'    => \$log_level,
           'f|lock-file=s'    => \$lock_file,
       );
       &printHelpAndExit() if ($help);

       $lock_file         = &DEFAULT_LOCK_FILE            if !defined $lock_file    || $lock_file    =~ /^\s*$/;
       $log_level         = &DEFAULT_LOG_LEVEL            if !defined $log_level    || $log_level    =~ /^\s*$/; 
       $socket_path       = &DEFAULT_SOCKET_PATH          if !defined $socket_path  || $socket_path  =~ /^\s*$/;
       $infected_dir      = &DEFAULT_PATH_TO_INFECTED_DIR if !defined $infected_dir || $infected_dir =~ /^\s*$/;

       $infected_dir =~ s{/$}{};

       &printHeaderDialog($socket_path, $infected_dir, $log_level, $lock_file);

       $logger = Logger->new($log_level);

       $logger->log('debug', "Locking: " . $lock_file);
       &lock($logger, $lock_file);

       $start_time = localtime();
       my WrapperSocket $socket = WrapperSocket->new(logger => $logger, peer_address => $socket_path);

       &File::Find::find(
           sub {
               die 'Interrapt processing...' if $exit_flag;
               eval {
                   if($File::Find::name !~ /\.envelope$/ || -d $File::Find::name) {
                       $logger->log('debug', "Founded file: '$File::Find::name'. Not envelope, skipped...");
                       return;
                   }

                   $logger->log('info', "Find envelope '$File::Find::name' to process.");
                   my Envelope $envelope = Envelope->new(logger => $logger, infected_dir => $infected_dir, env_file => $File::Find::name);

                   unless( $envelope->needToProcess() ) {
                       $logger->log('info', "Processing skipped. Reason: can not find message file for envelope '$File::Find::name'.");
                       $skipped_files++;
                       return;
                   }

                   $logger->log('info', "Parssing envelope '$File::Find::name'.");
                   my ($id, $from, $to) = ($envelope->getId(), $envelope->getFrom(), (join '>, <', @{$envelope->getTo()}));
                   $logger->log('info', "Sending request to add envelope '$File::Find::name'.");
                   $socket->sendData(sprintf "quarantine-add %s <%s> <%s>\n", $id, $from, $to);
                   $logger->log('info', "Receive answer...");
                   $output = '';
                   $socket->receiveData($output);
                   &checkAnswer($logger, $output);
                   $logger->log('info', "Deleting envelope '$File::Find::name'.");
                   $envelope->destroy();
                   $processed_files++;
               };
               if($@) {
                   if ( $@->isa('Exception::Envelope') ) {
                       $errors++;
                       $logger->log('warning', $@);
                   } else {
                       die $@;
                   }
               }
           }
           ,
           $infected_dir
       );

       $quit_flag = 1;
       $socket->sendData("quit");
       $socket->receiveData();# wait answer from maild
       $end_time = localtime();
   };
   if($@) {
       if( $@->isa('Exception::IO')) {
           $logger->log('error', "Exception::IO error: $@");
       } else {
           $logger->log('error', "Exception::Base: $@");
           if( defined $socket&& !$quit_flag ) {
               $socket->sendData("quit");
               $socket->receiveData();# wait answer from maild
           }
       }
   } 
   anycase: {
       &unlock($logger);
   }
   &printFotter($start_time, $end_time, $processed_files, $skipped_files, $errors);
}


{
    my $lock_file;
    my $fh;

    sub lock
    {
        my Logger $logger  = shift;
        $logger->log('debug', "Lock tring to file '$_[0]'");
        open( $fh, q{>>}, $_[0] )
            or die( "Can not open lockfile '$_[0]' for locking: $!" );
        flock( $fh, &Fcntl::LOCK_EX | &Fcntl::LOCK_NB )
            or die( "Can not obtain lockfile '$_[0]', stop to prevent multiply instances: $!" );
        $lock_file = $_[0];
        return 1;
    }


    sub unlock
    {
        my Logger $logger = shift;
        $logger->log('debug', 'Unlock requested.');
        if (defined $lock_file) {
            $logger->log('debug', "Unlock file '$lock_file'.");
            unlink($lock_file) or die "Can not unlink lock file: $!";
            flock( $fh, &Fcntl::LOCK_UN ) or die "Can not unlock lock file: $!";
            close $fh or die "Can not close file: $!";
            $lock_file = undef;
        } else {
            $logger->log('debug', 'Nothing to unlock.');
        }
        return 1;
    }
}


sub checkAnswer
{
    my Logger $logger = shift;
    $logger->log('debug', "Checking answer."); 
    if ($_[0] =~ m/^Error /) {
        die Exception::Envelope->new("Daemon returned error: " . $_[0]);
    }
    return 1;
}


sub printHeaderDialog
{
    my ($socket_path, $infected_dir, $log_level, $lock_file) = @_;

    print <<HEADER;
This program makes migration from quarantine version 4.44 to 5.00 .

Program will use following parameters:
    socket path       : $socket_path
    path ro quarantine: $infected_dir
    log level         : $log_level
    lockfile          : $lock_file
If you want to change parameters, you have to quit
the program and run command:
    $0 -h
to know how to do this.
HEADER
    print "If you want to start processing type [Yes/yes/y]: ";

    my $answer;
    eval {
        $SIG{INT}  = sub {print "\nCatched signal to quit.\n"; die "Canceled."};
        $answer = <>;
    };
    $answer =~ s/^\s*(.*?)\s*$/$1/ if defined $answer;
    $answer = lc $answer;
    if ( $answer eq 'yes' || $answer eq 'y' ) {
       print "Program has been started.\n";
    } else {
       die "Canceled.";
    }

    return 1;
}


sub printFotter
{
    my ($start_time, $end_time, $processed_files, $skipped_files, $errors) = @_;

    $start_time = localtime() unless defined $start_time;
    $end_time   = localtime() unless defined $end_time;

    print <<FOOTER;
Programm finished.
Start time                   : $start_time
End time                     : $end_time
Skipped envelopes            : $skipped_files
Processed envelopes          : $processed_files
Errors in processed envelopes: $errors
FOOTER

   return 1;
}


sub printHelpAndExit
{
    my ($DEFAULT_PATH_TO_INFECTED_DIR, $DEFAULT_SOCKET_PATH, $DEFAULT_LOCK_FILE)
         = (&DEFAULT_PATH_TO_INFECTED_DIR, &DEFAULT_SOCKET_PATH, &DEFAULT_LOCK_FILE);

    print <<HELP;
Ussage $0:
    $0 [<option1> <option2> ... <optionN>]

Options:
    q|infected-dir -- set path to quarantine (default path is '$DEFAULT_PATH_TO_INFECTED_DIR')
    s|socket-path  -- set path to socket (defualt path is '$DEFAULT_SOCKET_PATH')
    l|log-level    -- log level (available levels: quiet, error, warning, info, debug)
    f|lock-file    -- set name of lockfile (default path is '$DEFAULT_LOCK_FILE')
    h|help         -- print this help & exit

Example:
    $0 -h
    $0 -s=/var/drweb/ipc/.ctl
    $0 -q=/var/drweb/infected -s=/var/drweb/ipc/.ctl 
    $0 -q=/var/drweb/infected -l=debug
HELP
    exit 0;
}
